@@ -23,15 +23,11 @@ import subprocess
 from ..mesonlib import MesonException, OrderedSet
 from ..mesonlib import classify_unity_sources
 from ..mesonlib import File
-from ..compilers import CompilerArgs, get_macos_dylib_install_name
+from ..compilers import CompilerArgs, VisualStudioCCompiler
 from collections import OrderedDict
 import shlex
 from functools import lru_cache
 
-@lru_cache(maxsize=None)
-def get_target_macos_dylib_install_name(ld):
-    return get_macos_dylib_install_name(ld.prefix, ld.name, ld.suffix, ld.soversion)
-
 
 class CleanTrees:
     '''
@@ -135,8 +131,8 @@ class Backend:
         self.build = build
         self.environment = build.environment
         self.processed_targets = {}
-        self.build_to_src = os.path.relpath(self.environment.get_source_dir(),
-                                            self.environment.get_build_dir())
+        self.build_to_src = mesonlib.relpath(self.environment.get_source_dir(),
+                                             self.environment.get_build_dir())
 
     def get_target_filename(self, t):
         if isinstance(t, build.CustomTarget):
@@ -215,6 +211,7 @@ class Backend:
     def get_target_private_dir_abs(self, target):
         return os.path.join(self.environment.get_build_dir(), self.get_target_private_dir(target))
 
+    @lru_cache(maxsize=None)
     def get_target_generated_dir(self, target, gensrc, src):
         """
         Takes a BuildTarget, a generator source (CustomTarget or GeneratedList),
@@ -446,7 +443,13 @@ class Backend:
                 sources.append(File(True, dirpart, fnamepart))
 
         # Filter out headers and all non-source files
-        sources = [s for s in sources if self.environment.is_source(s) and not self.environment.is_header(s)]
+        filtered_sources = []
+        for s in sources:
+            if self.environment.is_source(s) and not self.environment.is_header(s):
+                filtered_sources.append(s)
+            elif self.environment.is_object(s):
+                result.append(s.relative_name())
+        sources = filtered_sources
 
         # extobj could contain only objects and no sources
         if not sources:
@@ -488,7 +491,7 @@ class Backend:
             return args
         extra_args = []
         # Compiler-specific escaping is needed for -D args but not for any others
-        if compiler.get_id() == 'msvc':
+        if isinstance(compiler, VisualStudioCCompiler):
             # MSVC needs escaping when a -D argument ends in \ or \"
             for arg in args:
                 if arg.startswith('-D') or arg.startswith('/D'):
@@ -555,6 +558,8 @@ class Backend:
         # Set -fPIC for static libraries by default unless explicitly disabled
         if isinstance(target, build.StaticLibrary) and target.pic:
             commands += compiler.get_pic_args()
+        if isinstance(target, build.Executable) and target.pie:
+            commands += compiler.get_pie_args()
         # Add compile args needed to find external dependencies. Link args are
         # added while generating the link command.
         # NOTE: We must preserve the order in which external deps are
@@ -602,6 +607,8 @@ class Backend:
                 continue
             if compiler.get_language() == 'd':
                 arg = '-Wl,' + arg
+            else:
+                arg = compiler.get_linker_lib_prefix() + arg
             args.append(arg)
         return args
 
@@ -972,61 +979,11 @@ class Backend:
         with open(install_data_file, 'wb') as ofile:
             pickle.dump(d, ofile)
 
-    def get_target_install_dirs(self, t):
-        # Find the installation directory.
-        if isinstance(t, build.SharedModule):
-            default_install_dir = self.environment.get_shared_module_dir()
-        elif isinstance(t, build.SharedLibrary):
-            default_install_dir = self.environment.get_shared_lib_dir()
-        elif isinstance(t, build.StaticLibrary):
-            default_install_dir = self.environment.get_static_lib_dir()
-        elif isinstance(t, build.Executable):
-            default_install_dir = self.environment.get_bindir()
-        elif isinstance(t, build.CustomTarget):
-            default_install_dir = None
-        else:
-            assert(isinstance(t, build.BuildTarget))
-            # XXX: Add BuildTarget-specific install dir cases here
-            default_install_dir = self.environment.get_libdir()
-        outdirs = t.get_custom_install_dir()
-        if outdirs[0] is not None and outdirs[0] != default_install_dir and outdirs[0] is not True:
-            # Either the value is set to a non-default value, or is set to
-            # False (which means we want this specific output out of many
-            # outputs to not be installed).
-            custom_install_dir = True
-        else:
-            custom_install_dir = False
-            outdirs[0] = default_install_dir
-        return outdirs, custom_install_dir
-
-    def get_target_link_deps_mappings(self, t, prefix):
-        '''
-        On macOS, we need to change the install names of all built libraries
-        that a target depends on using install_name_tool so that the target
-        continues to work after installation. For this, we need a dictionary
-        mapping of the install_name value to the new one, so we can change them
-        on install.
-        '''
-        result = {}
-        if isinstance(t, build.StaticLibrary):
-            return result
-        for ld in t.get_all_link_deps():
-            if ld is t or not isinstance(ld, build.SharedLibrary):
-                continue
-            old = get_target_macos_dylib_install_name(ld)
-            if old in result:
-                continue
-            fname = ld.get_filename()
-            outdirs, _ = self.get_target_install_dirs(ld)
-            new = os.path.join(prefix, outdirs[0], fname)
-            result.update({old: new})
-        return result
-
     def generate_target_install(self, d):
         for t in self.build.get_targets().values():
             if not t.should_install():
                 continue
-            outdirs, custom_install_dir = self.get_target_install_dirs(t)
+            outdirs, custom_install_dir = t.get_install_dir(self.environment)
             # Sanity-check the outputs and install_dirs
             num_outdirs, num_out = len(outdirs), len(t.get_outputs())
             if num_outdirs != 1 and num_outdirs != num_out:
@@ -1041,7 +998,7 @@ class Backend:
                 # Install primary build output (library/executable/jar, etc)
                 # Done separately because of strip/aliases/rpath
                 if outdirs[0] is not False:
-                    mappings = self.get_target_link_deps_mappings(t, d.prefix)
+                    mappings = t.get_link_deps_mapping(d.prefix, self.environment)
                     i = TargetInstallData(self.get_target_filename(t), outdirs[0],
                                           t.get_aliases(), should_strip, mappings,
                                           t.install_rpath, install_mode)
@@ -1136,7 +1093,7 @@ class Backend:
                 if subdir is None:
                     subdir = os.path.join(manroot, 'man' + num)
                 srcabs = f.absolute_path(self.environment.get_source_dir(), self.environment.get_build_dir())
-                dstabs = os.path.join(subdir, os.path.basename(f.fname) + '.gz')
+                dstabs = os.path.join(subdir, os.path.basename(f.fname))
                 i = [srcabs, dstabs, m.get_custom_install_mode()]
                 d.man.append(i)
 
